@@ -1,5 +1,5 @@
 skip_if_no_token <- function() {
-  if (is.na(Sys.getenv("GH_TESTING", NA_character_))) {
+  if (Sys.getenv("GH_TESTING") == "") {
     skip("No GitHub token")
   }
 }
